@@ -2,15 +2,19 @@ package workloadcontroller
 
 import (
 	"fmt"
+	"reflect"
 	"sort"
 	"strings"
 
+	"github.com/ghodss/yaml"
+
 	appsv1 "k8s.io/api/apps/v1"
 	corev1 "k8s.io/api/core/v1"
 	"k8s.io/apimachinery/pkg/api/equality"
 	apierrors "k8s.io/apimachinery/pkg/api/errors"
 	"k8s.io/apimachinery/pkg/api/meta"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
+	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
 	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/apimachinery/pkg/runtime/schema"
 	"k8s.io/apimachinery/pkg/util/sets"
@@ -22,6 +26,7 @@ import (
 	apiregistrationv1client "k8s.io/kube-aggregator/pkg/client/clientset_generated/clientset/typed/apiregistration/v1"
 
 	openshiftapi "github.com/openshift/api"
+	openshiftcontrolplanev1 "github.com/openshift/api/openshiftcontrolplane/v1"
 	operatorv1 "github.com/openshift/api/operator/v1"
 	openshiftconfigclientv1 "github.com/openshift/client-go/config/clientset/versioned/typed/config/v1"
 	"github.com/openshift/cluster-openshift-apiserver-operator/pkg/operator/operatorclient"
@@ -77,6 +82,11 @@ func syncOpenShiftAPIServer_v311_00_to_latest(c OpenShiftAPIServerOperator, orig
 		reasonsForForcedRollingUpdate = append(reasonsForForcedRollingUpdate, "modified: "+resourceSelectorForCLI(imageImportCAModifiedObject))
 	}
 
+	err = syncOpenShiftAPIServerTrustedCA_v311_00_to_latest(c.kubeClient.CoreV1(), c.eventRecorder)
+	if err != nil {
+		errors = append(errors, fmt.Errorf("%q: %v", "trusted-ca-bundle", err))
+	}
+
 	if operatorConfig.ObjectMeta.Generation != operatorConfig.Status.ObservedGeneration {
 		reasonsForForcedRollingUpdate = append(reasonsForForcedRollingUpdate, fmt.Sprintf("operator config spec generation %d does not match status generation %d", operatorConfig.ObjectMeta.Generation,
 			operatorConfig.Status.ObservedGeneration))
@@ -91,7 +101,7 @@ func syncOpenShiftAPIServer_v311_00_to_latest(c OpenShiftAPIServerOperator, orig
 
 	// our configmaps and secrets are in order, now it is time to create the DS
 	// TODO check basic preconditions here
-	actualDaemonSet, _, err := manageOpenShiftAPIServerDaemonSet_v311_00_to_latest(c.kubeClient.AppsV1(), c.eventRecorder, c.targetImagePullSpec, operatorConfig, operatorConfig.Status.Generations, forceRollingUpdate)
+	actualDaemonSet, _, err := manageOpenShiftAPIServerDaemonSet_v311_00_to_latest(c.kubeClient.AppsV1(), c.eventRecorder, c.targetImagePullSpec, c.operatorImagePullSpec, operatorConfig, operatorConfig.Status.Generations, forceRollingUpdate)
 	if err != nil {
 		errors = append(errors, fmt.Errorf("%q: %v", "daemonsets", err))
 	}
@@ -135,7 +145,7 @@ func syncOpenShiftAPIServer_v311_00_to_latest(c OpenShiftAPIServerOperator, orig
 	// if the apiservices themselves check out ok, try to actually hit the discovery endpoints.  We have a history in clusterup
 	// of something delaying them.  This isn't perfect because of round-robining, but let's see if we get an improvement
 	if len(availableConditionMessages) == 0 && c.kubeClient.Discovery().RESTClient() != nil {
-		missingAPIMessages := checkForAPIs(c.kubeClient.Discovery().RESTClient(), apiServiceGroupVersions...)
+		missingAPIMessages := checkForAPIs(c.eventRecorder, c.kubeClient.Discovery().RESTClient(), apiServiceGroupVersions...)
 		availableConditionMessages = append(availableConditionMessages, missingAPIMessages...)
 	}
 
@@ -273,10 +283,59 @@ func manageOpenShiftAPIServerImageImportCA_v311_00_to_latest(openshiftConfigClie
 	}
 }
 
+func syncOpenShiftAPIServerTrustedCA_v311_00_to_latest(client coreclientv1.CoreV1Interface, recorder events.Recorder) error {
+	// get the CM created by the CVO in the operator namespace
+	sourceCM, err := client.ConfigMaps(operatorclient.OperatorNamespace).Get("trusted-ca-bundle", metav1.GetOptions{})
+	if err != nil {
+		return err
+	}
+
+	// it's optional, don't sync it if it has no data -> prevent trust store being removed upon startup/upgrade
+	if len(sourceCM.Data) == 0 {
+		return nil
+	}
+
+	// get the copy of the sourceCM, remove the annotations and labels we set for CVO and network operator
+	requiredCM := sourceCM.DeepCopy()
+	requiredCM.UID = ""
+	requiredCM.ObjectMeta.ResourceVersion = ""
+	requiredCM.ObjectMeta.Namespace = operatorclient.TargetNamespace
+	delete(requiredCM.Annotations, "release.openshift.io/create-only")
+	delete(requiredCM.Labels, "config.openshift.io/inject-trusted-cabundle")
+
+	targetClient := client.ConfigMaps(operatorclient.TargetNamespace)
+	imageCM, err := targetClient.Get(requiredCM.Name, metav1.GetOptions{})
+	if apierrors.IsNotFound(err) {
+		_, err = targetClient.Create(requiredCM)
+		return err
+	}
+
+	if err != nil {
+		return err
+	}
+
+	// if the target CM Data differs, update it
+	if !reflect.DeepEqual(sourceCM.Data, imageCM.Data) {
+		requiredCM.UID = imageCM.UID
+		requiredCM.ResourceVersion = imageCM.ResourceVersion
+		_, err = targetClient.Update(requiredCM)
+	}
+
+	return err
+}
+
 func manageOpenShiftAPIServerConfigMap_v311_00_to_latest(kubeClient kubernetes.Interface, client coreclientv1.ConfigMapsGetter, recorder events.Recorder, operatorConfig *operatorv1.OpenShiftAPIServer) (*corev1.ConfigMap, bool, error) {
 	configMap := resourceread.ReadConfigMapV1OrDie(v311_00_assets.MustAsset("v3.11.0/openshift-apiserver/cm.yaml"))
 	defaultConfig := v311_00_assets.MustAsset("v3.11.0/openshift-apiserver/defaultconfig.yaml")
-	requiredConfigMap, _, err := resourcemerge.MergeConfigMap(configMap, "config.yaml", nil, defaultConfig, operatorConfig.Spec.ObservedConfig.Raw, operatorConfig.Spec.UnsupportedConfigOverrides.Raw)
+	requiredConfigMap, _, err := resourcemerge.MergePrunedConfigMap(
+		&openshiftcontrolplanev1.OpenShiftAPIServerConfig{},
+		configMap,
+		"config.yaml",
+		nil,
+		defaultConfig,
+		operatorConfig.Spec.ObservedConfig.Raw,
+		operatorConfig.Spec.UnsupportedConfigOverrides.Raw,
+	)
 	if err != nil {
 		return nil, false, err
 	}
@@ -286,6 +345,7 @@ func manageOpenShiftAPIServerConfigMap_v311_00_to_latest(kubeClient kubernetes.I
 		kubeClient,
 		resourcehash.NewObjectRef().ForSecret().InNamespace(operatorclient.TargetNamespace).Named("etcd-client"),
 		resourcehash.NewObjectRef().ForConfigMap().InNamespace(operatorclient.TargetNamespace).Named("etcd-serving-ca"),
+		resourcehash.NewObjectRef().ForConfigMap().InNamespace(operatorclient.TargetNamespace).Named("trusted-ca-bundle"),
 	)
 	if err != nil {
 		return nil, false, err
@@ -297,14 +357,43 @@ func manageOpenShiftAPIServerConfigMap_v311_00_to_latest(kubeClient kubernetes.I
 	return resourceapply.ApplyConfigMap(client, recorder, requiredConfigMap)
 }
 
-func manageOpenShiftAPIServerDaemonSet_v311_00_to_latest(client appsclientv1.DaemonSetsGetter, recorder events.Recorder, imagePullSpec string, operatorConfig *operatorv1.OpenShiftAPIServer, generationStatus []operatorv1.GenerationStatus, forceRollingUpdate bool) (*appsv1.DaemonSet, bool, error) {
+func manageOpenShiftAPIServerDaemonSet_v311_00_to_latest(
+	client appsclientv1.DaemonSetsGetter,
+	recorder events.Recorder,
+	imagePullSpec string,
+	operatorImagePullSpec string,
+	operatorConfig *operatorv1.OpenShiftAPIServer,
+	generationStatus []operatorv1.GenerationStatus,
+	forceRollingUpdate bool) (*appsv1.DaemonSet, bool, error) {
 	required := resourceread.ReadDaemonSetV1OrDie(v311_00_assets.MustAsset("v3.11.0/openshift-apiserver/ds.yaml"))
-	if len(imagePullSpec) > 0 {
-		required.Spec.Template.Spec.Containers[0].Image = imagePullSpec
-		if len(required.Spec.Template.Spec.InitContainers) > 0 {
-			required.Spec.Template.Spec.InitContainers[0].Image = imagePullSpec
+
+	for i := range required.Spec.Template.Spec.Containers {
+		img := strings.TrimSpace(required.Spec.Template.Spec.Containers[i].Image)
+		switch img {
+		case "${IMAGE}":
+			required.Spec.Template.Spec.Containers[i].Image = imagePullSpec
+		case "${OPERATOR_IMAGE}":
+			required.Spec.Template.Spec.Containers[i].Image = operatorImagePullSpec
+		default:
+			if strings.Contains(img, "$") {
+				return nil, false, fmt.Errorf("invalid image reference %q", img)
+			}
 		}
 	}
+	for i := range required.Spec.Template.Spec.InitContainers {
+		img := strings.TrimSpace(required.Spec.Template.Spec.InitContainers[i].Image)
+		switch img {
+		case "${IMAGE}":
+			required.Spec.Template.Spec.InitContainers[i].Image = imagePullSpec
+		case "${OPERATOR_IMAGE}":
+			required.Spec.Template.Spec.InitContainers[i].Image = operatorImagePullSpec
+		default:
+			if strings.Contains(img, "$") {
+				return nil, false, fmt.Errorf("invalid image reference %q", img)
+			}
+		}
+	}
+
 	// we set this so that when the requested image pull spec changes, we always have a diff.  Remember that we don't directly
 	// diff any fields on the daemonset because they can be rewritten by admission and we don't want to constantly be fighting
 	// against admission or defaults.  That was a problem with original versions of apply.
@@ -312,18 +401,42 @@ func manageOpenShiftAPIServerDaemonSet_v311_00_to_latest(client appsclientv1.Dae
 		required.Annotations = map[string]string{}
 	}
 	required.Annotations["openshiftapiservers.operator.openshift.io/pull-spec"] = imagePullSpec
+	required.Annotations["openshiftapiservers.operator.openshift.io/operator-pull-spec"] = operatorImagePullSpec
 
+	containerArgsWithLoglevel := required.Spec.Template.Spec.Containers[0].Args
+	if argsCount := len(containerArgsWithLoglevel); argsCount > 1 {
+		return nil, false, fmt.Errorf("expected only one container argument, got %d", argsCount)
+	}
+	if !strings.Contains(containerArgsWithLoglevel[0], "exec openshift-apiserver") {
+		return nil, false, fmt.Errorf("exec openshift-apiserver not found in first argument %q", containerArgsWithLoglevel[0])
+	}
+	containerArgsWithLoglevel[0] = strings.TrimSpace(containerArgsWithLoglevel[0])
 	switch operatorConfig.Spec.LogLevel {
 	case operatorv1.Normal:
-		required.Spec.Template.Spec.Containers[0].Args = append(required.Spec.Template.Spec.Containers[0].Args, fmt.Sprintf("-v=%d", 2))
+		containerArgsWithLoglevel[0] += fmt.Sprintf(" -v=%d", 2)
 	case operatorv1.Debug:
-		required.Spec.Template.Spec.Containers[0].Args = append(required.Spec.Template.Spec.Containers[0].Args, fmt.Sprintf("-v=%d", 4))
+		containerArgsWithLoglevel[0] += fmt.Sprintf(" -v=%d", 4)
 	case operatorv1.Trace:
-		required.Spec.Template.Spec.Containers[0].Args = append(required.Spec.Template.Spec.Containers[0].Args, fmt.Sprintf("-v=%d", 6))
+		containerArgsWithLoglevel[0] += fmt.Sprintf(" -v=%d", 6)
 	case operatorv1.TraceAll:
-		required.Spec.Template.Spec.Containers[0].Args = append(required.Spec.Template.Spec.Containers[0].Args, fmt.Sprintf("-v=%d", 8))
+		containerArgsWithLoglevel[0] += fmt.Sprintf(" -v=%d", 8)
 	default:
-		required.Spec.Template.Spec.Containers[0].Args = append(required.Spec.Template.Spec.Containers[0].Args, fmt.Sprintf("-v=%d", 2))
+		containerArgsWithLoglevel[0] += fmt.Sprintf(" -v=%d", 2)
+	}
+	required.Spec.Template.Spec.Containers[0].Args = containerArgsWithLoglevel
+
+	var observedConfig map[string]interface{}
+	if err := yaml.Unmarshal(operatorConfig.Spec.ObservedConfig.Raw, &observedConfig); err != nil {
+		return nil, false, fmt.Errorf("failed to unmarshal the observedConfig: %v", err)
+	}
+	proxyConfig, _, err := unstructured.NestedStringMap(observedConfig, "workloadcontroller", "proxy")
+	if err != nil {
+		return nil, false, fmt.Errorf("couldn't get the proxy config from observedConfig: %v", err)
+	}
+
+	proxyEnvVars := proxyMapToEnvVars(proxyConfig)
+	for i, container := range required.Spec.Template.Spec.Containers {
+		required.Spec.Template.Spec.Containers[i].Env = append(container.Env, proxyEnvVars...)
 	}
 
 	return resourceapply.ApplyDaemonSet(client, recorder, required, resourcemerge.ExpectedDaemonSetGeneration(required, generationStatus), forceRollingUpdate)
@@ -397,3 +510,18 @@ func resourceSelectorForCLI(obj runtime.Object) string {
 	}
 	return kind + group + "/" + name
 }
+
+func proxyMapToEnvVars(proxyConfig map[string]string) []corev1.EnvVar {
+	if proxyConfig == nil {
+		return nil
+	}
+
+	envVars := []corev1.EnvVar{}
+	for k, v := range proxyConfig {
+		envVars = append(envVars, corev1.EnvVar{Name: k, Value: v})
+	}
+
+	// sort the env vars to prevent update hotloops
+	sort.Slice(envVars, func(i, j int) bool { return envVars[i].Name < envVars[j].Name })
+	return envVars
+}
